#include "screen.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

static int _8_16[256];
static int _8_32[256];

void set_colormap(SCREEN *s)
{
  XColor color[256];
  int i;

  if (s->depth!=8)
    return;

  for (i=0; i<256; i++) {
    color[i].pixel=i;
    color[i].flags=DoRed|DoGreen|DoBlue;
    color[i].red=(float)i/255.*65535;
    color[i].green=(float)i/255.*65535;
    color[i].blue=(float)i/255.*65535;
  }
  XStoreColors(s->d, s->cm, color, 256);
}

int to16(int col)
{
  return _8_16[col];
}

int to32(int col)
{
  return _8_32[col];
}

/* return -1 if error, 0 if ok */
int init_screen(SCREEN *s)
{
  XSetWindowAttributes attributs;
  XGCValues gcv;
  XEvent ev;
  XColor a;
  Pixmap p;
  GC pixgc;
  int i;

  for (i=0; i<256; i++) {
    _8_16[i]=((i>>3)<<(5+6))+((i>>2)<<5)+(i>>3);
  }

  for (i=0; i<256; i++) {
    int r, g, b;
    r=i;
    g=i;
    b=i;
    _8_32[i]=(r<<16)|(g<<8)|b;
  }

  memset(s, 0, sizeof(SCREEN));

  if (!(s->d=XOpenDisplay((char *)0)))
    return -1;

  s->depth=DefaultDepth(s->d, DefaultScreen(s->d));

  if (s->depth!=8 && s->depth!=16 && s->depth!=24) {
    fprintf(stderr, "screen depth not supported (only 8, 16 and 24bpp (which means 32bpp too) handled\n");
    XCloseDisplay(s->d);
    s->d=(Display *)0;
    return -1;
  }

  /* we don't know if s->depth==24 if it's 24 or 32bpp, let's check it */
  if (s->depth==24) {
    int n;
    XPixmapFormatValues *v;

    v=XListPixmapFormats(s->d, &n);
    if (!v || !n) {
      if (v)
	XFree(v);
      fprintf(stderr, "Error while trying to distinguish between 24 or 32bpp, damned...\n");
      XCloseDisplay(s->d);
      s->d=(Display *)0;
      return -1;
    }
    for (n--; n>=0; n--)
      if (v[n].bits_per_pixel==32)
	s->depth=32;
    XFree(v);
  }

  s->buffer=(char *)malloc(320*240/8*s->depth);
  if (!s->buffer) {
    perror("");
    fprintf(stderr, "Meaning no rubix\n");
    XCloseDisplay(s->d);
    s->d=(Display *)0;
    return -1;
  }

  if (s->depth==8) {
    s->cm=XCreateColormap(s->d, DefaultRootWindow(s->d), DefaultVisual(s->d, DefaultScreen(s->d)), AllocAll);
    attributs.colormap=s->cm;
  }
  /* let's crate a void cursor for our window */
  p=XCreatePixmap(s->d, DefaultRootWindow(s->d), 1, 1, 1);
  pixgc=XCreateGC(s->d, p, 0, 0);
  a.red=a.green=a.blue=0;
  XFillRectangle(s->d, p, pixgc, 0, 0, 2, 2);
  XFreeGC(s->d, pixgc);
  attributs.cursor=XCreatePixmapCursor(s->d, p, p, &a, &a, 0, 0);
  XFreePixmap(s->d, p);
  /* let's create and map our window */
  if (s->depth==8) {
    s->w=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 320, 240, 3, 8, CopyFromParent, CopyFromParent,
		       CWColormap|CWCursor, &attributs);
  } else {
    s->w=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 320, 240, 3, s->depth==32?24:s->depth,
		       CopyFromParent, CopyFromParent, CWCursor, &attributs);
  }
  XSelectInput(s->d, s->w, KeyPressMask|ExposureMask|PointerMotionMask/*|FocusChangeMask*/|ButtonPressMask);
  XMapWindow(s->d, s->w);
  do
    XNextEvent(s->d, &ev);
  while (ev.type!=Expose || ev.xexpose.window!=s->w);
  gcv.function=GXcopy;
  s->gc=XCreateGC(s->d,s->w,GCFunction,&gcv);
  /* create our XImage */
  /* bitmap_pad, what the heck is it ?? mail sed@free.fr if you know, please...
   * Well, seems it has to be 32...
   */
  s->im=XCreateImage(s->d, DefaultVisual(s->d, DefaultScreen(s->d)), s->depth==32?24:s->depth, ZPixmap,
		     0, s->buffer, 320, 240, 32, 0);
  if (s->depth==8)
    set_colormap(s);
  return 0;
}

void close_screen(SCREEN *s)
{
  XDestroyWindow(s->d, s->w);
  XFreeGC(s->d, s->gc);
  if (s->depth==8)
    XFreeColormap(s->d, s->cm);
  XCloseDisplay(s->d);
}

void clear_screen(SCREEN *s)
{
  memset(s->buffer, 0, 320*240/8*s->depth);
}

void put_screen(SCREEN *s)
{
  extern unsigned char *ecran1;
  /* convertion */
  if (s->depth==8)
    memcpy(s->buffer, ecran1, 320*240);
  if (s->depth==16) {
    register unsigned char *p=ecran1;
    register unsigned short *q=(unsigned short *)s->buffer;
    register int i;
    for (i=320*240; i>0; i--, q++, p++)
      *q=_8_16[*p];
  }
  /* bullshit code this 8->24 converter, nope ? */
  if (s->depth==24) {
    register unsigned char *p=ecran1;
    register unsigned char *q=(unsigned char *)s->buffer;
    register int i;
    for (i=320*240; i>0; i--, q++, p++) {
      int d=_8_32[*p];
      *q++=(d>>16)&255;
      *q++=(d>>8)&255;
      *q=d&255;
    }
  }
  if (s->depth==32) {
    register unsigned char *p=ecran1;
    register unsigned int *q=(unsigned int *)s->buffer;
    register int i;
    for (i=320*240; i>0; i--, q++, p++)
      *q=_8_32[*p];
  }
  XPutImage(s->d, s->w, s->gc, s->im, 0, 0, 0, 0, 320, 240);
  XFlush(s->d);
}
